@@ -1,32 +1,3 @@
-
-/* cgeneric.h
- * 
- * Copyright (C) 2021-2024 Havard Rue
- * 
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or (at
- * your option) any later version.
- * 
- * This program is distributed in the hope that it will be useful, but
- * WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * General Public License for more details.
- * 
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
- *
- * The author's contact information:
- *
- *        Haavard Rue
- *        CEMSE Division
- *        King Abdullah University of Science and Technology
- *        Thuwal 23955-6900, Saudi Arabia
- *        Email: haavard.rue@kaust.edu.sa
- *        Office: +966 (0)12 808 0640
- *
- */
 #ifndef __INLA_CGENERIC_H__
 #define __INLA_CGENERIC_H__
 
@@ -40,10 +11,23 @@
 #define __END_DECLS					       /* empty */
 #endif
 __BEGIN_DECLS
+#include <assert.h>
+#include <math.h>
+#include <stdio.h>
+#include <omp.h>
+#include <stdio.h>
 
 /* 
  *
  */
+// same definition as in GMRFLibP.h...
+#if !defined(POSSIBLY_UNUSED)
+#ifdef __GNUC__
+#define POSSIBLY_UNUSED(x) __attribute__((__unused__)) x
+#else
+#define POSSIBLY_UNUSED(x) x
+#endif
+#endif
     typedef enum {
 	INLA_CGENERIC_VOID = 0,
 	INLA_CGENERIC_Q,
@@ -55,14 +39,27 @@ __BEGIN_DECLS
 	INLA_CGENERIC_QUIT
 } inla_cgeneric_cmd_tp;
 
-#define INLA_CGENERIC_CMD_NAME(cmd_) ((cmd_) == INLA_CGENERIC_VOID ? "void" : \
-				      ((cmd_) == INLA_CGENERIC_Q ? "Q" : \
+typedef enum {
+	INLA_CLOGLIKE_INITIAL = 1,
+	INLA_CLOGLIKE_LOG_PRIOR,
+	INLA_CLOGLIKE_LOGLIKE,
+	INLA_CLOGLIKE_CDF,
+	INLA_CLOGLIKE_QUIT
+} inla_cloglike_cmd_tp;
+
+#define INLA_CGENERIC_CMD_NAME(cmd_) ((cmd_) == INLA_CGENERIC_Q ? "Q" : \
 				       ((cmd_) == INLA_CGENERIC_GRAPH ? "graph" : \
 					((cmd_) == INLA_CGENERIC_MU ? "mu" : \
 					 ((cmd_) == INLA_CGENERIC_INITIAL ? "initial" : \
 					  ((cmd_) == INLA_CGENERIC_LOG_NORM_CONST ? "log_norm_const" : \
 					   ((cmd_) == INLA_CGENERIC_LOG_PRIOR ? "log_prior" : \
-					    ((cmd_) == INLA_CGENERIC_QUIT ? "quit" : "(***ERROR***)"))))))))
+					    ((cmd_) == INLA_CGENERIC_QUIT ? "quit" : "(***ERROR***)")))))))
+
+#define INLA_CLOGLIKE_CMD_NAME(cmd_) ((cmd_) == INLA_CLOGLIKE_INITIAL ? "initial" : \
+				      ((cmd_) == INLA_CLOGLIKE_LOG_PRIOR ? "log_prior" : \
+				       ((cmd_) == INLA_CLOGLIKE_LOGLIKE ? "log_like" : \
+					((cmd_) == INLA_CLOGLIKE_LOGLIKE ? "CDF" : \
+					 ((cmd_) == INLA_CLOGLIKE_QUIT ? "quit" : "(***ERROR***)")))))
 
 /*
  *      matrix storage is stored row by row.
@@ -170,36 +167,71 @@ typedef struct {
 	int n_smats;
 	inla_cgeneric_smat_tp **smats;
 
+	int processed;
 	void *cache;
 } inla_cgeneric_data_tp;
 
 #if defined(_OPENMP)
 // tools useful for creating a cache
-#include <omp.h>
 #define IMAX_(a_,  b_) ((a_) >= (b_) ? (a_) : (b_))
-#define MAX_THREADS(data_) ((data_)->max_threads)
-#define CGENERIC_CACHE_LEN(data_) (IMAX_(1, MAX_THREADS(data_)) * (IMAX_(1, MAX_THREADS(data_)) + 1))
+#define MAX_THREADS_(data_) IMAX_(1, ((data_)->threads.max))
+#define CGENERIC_CACHE_LEN(data_) (MAX_THREADS_(data_) * (MAX_THREADS_(data_) + 1))
 #define CGENERIC_CACHE_ASSIGN_IDX(idx_, data_)				\
-        if (1) {                                                        \
-                int level_ = omp_get_level();                           \
-                int tnum_ = omp_get_thread_num();                       \
-                if (level_ <= 1)        {                               \
-                        idx_ =  tnum_;                                  \
-                } else if (level_ == 2) {                               \
-                        int level2_ = omp_get_ancestor_thread_num(level_ -1); \
-			idx_ = IMAX_(1, 1 + level2_) * IMAX_(1, MAX_THREADS(data_)) + tnum_; \
-                } else {                                                \
-                        assert(0 == 1);                                 \
-                }                                                       \
-        }
+        {								\
+		int level1_ = omp_get_level();				\
+		int tnum1_ = omp_get_thread_num();			\
+		if (level1_ <= 1) {					\
+			idx_ =  tnum1_;					\
+		} else if (level1_ == 2) {				\
+			int mt_ = MAX_THREADS_(data_);			\
+			int tnum2_ = omp_get_ancestor_thread_num(level1_ -1); \
+			idx_ = mt_ + tnum1_ + tnum2_ * mt_;		\
+		} else {						\
+			assert(0 == 1);					\
+		}							\
+	}
 #endif
 
 typedef double *inla_cgeneric_func_tp(inla_cgeneric_cmd_tp cmd, double *theta, inla_cgeneric_data_tp * data);
+typedef double *inla_cloglike_func_tp(inla_cloglike_cmd_tp cmd, double *theta, inla_cgeneric_data_tp * data,
+				      int ny, double *y, int nx, double *x, double *result);
 
 inla_cgeneric_data_tp *inla_cgeneric_read_data(const char *filename, int debug);
 inla_cgeneric_func_tp inla_cgeneric_iid_model;
 inla_cgeneric_func_tp inla_cgeneric_ar1_model;
 inla_cgeneric_func_tp inla_cgeneric_generic0_model;
 
+static void POSSIBLY_UNUSED(inla_cgeneric_data_print) (FILE * fp, inla_cgeneric_data_tp * data) {
+	fp = (fp ? fp : stdout);
+
+	fprintf(fp, "\nContent of cgeneric_data\n");
+	fprintf(fp, "\tthread.max   = [%1d]\n", data->threads.max);
+	fprintf(fp, "\tthread.outer = [%1d]\n", data->threads.outer);
+	fprintf(fp, "\tthread.inner = [%1d]\n", data->threads.inner);
+
+	fprintf(fp, "\tnumber of ints = [%1d]\n", data->n_ints);
+	for (int i = 0; i < data->n_ints; i++) {
+		fprintf(fp, "\t\tname[%s] length[%1d]\n", data->ints[i]->name, data->ints[i]->len);
+	}
+	fprintf(fp, "\tnumber of doubles = [%1d]\n", data->n_doubles);
+	for (int i = 0; i < data->n_doubles; i++) {
+		fprintf(fp, "\t\tname[%s] length[%1d]\n", data->doubles[i]->name, data->doubles[i]->len);
+	}
+	fprintf(fp, "\tnumber of chars = [%1d]\n", data->n_chars);
+	for (int i = 0; i < data->n_chars; i++) {
+		fprintf(fp, "\t\tname[%s] length[%1d] value[%s]\n", data->chars[i]->name, data->chars[i]->len, data->chars[i]->chars);
+	}
+	fprintf(fp, "\tnumber of matrices = [%1d]\n", data->n_mats);
+	for (int i = 0; i < data->n_mats; i++) {
+		fprintf(fp, "\t\tname[%s] dimension[%1d x %1d]\n", data->mats[i]->name, data->mats[i]->nrow, data->mats[i]->ncol);
+	}
+	fprintf(fp, "\tnumber of sparse matrices = [%1d]\n", data->n_smats);
+	for (int i = 0; i < data->n_smats; i++) {
+		fprintf(fp, "\t\tname[%s] dimension[%1d x %1d] nelements[%1d]\n", data->smats[i]->name, data->smats[i]->nrow, data->smats[i]->ncol,
+			data->smats[i]->n);
+	}
+	fprintf(fp, "\n");
+}
+
 __END_DECLS
 #endif
